#include <cmath>
#include <sstream>
#include <iomanip>

#include "XMLException.h"
#include "Configuration.h"
#include "Tiles.h"

#include "ObjectRepository.h"
#include "AllObjects.h"
#include "GameControlBase.h"

#include "PlayGroundVisitors.h"
#include "PlayGround.h"


//----------------------------------------------------------------------------
PlayGround *PlayGround::sm_instance = NULL;


//----------------------------------------------------------------------------
void CreateGravityVisitor::do_visit(const XMLNode *n)
{
    const std::string &name = n->getName();
    PlayGround *playGround = PlayGround::getInstance();

    if (name == "gravity")
    {
        n->acceptAllChilds(*this);
    }
    else if (name == "global")
    {
        int gx = n->getIntProperty("gx", 0);
        int gy = n->getIntProperty("gy", 100);

        int vx = n->getIntProperty("vx", 0);
        int vy = n->getIntProperty("vy", 0);

        unsigned friction = n->getUnsignedProperty("friction", 50);

        playGround->setXGravityAll(gx);
        playGround->setYGravityAll(gy);
        playGround->setXVelocityAll(vx);
        playGround->setYVelocityAll(vy);
        playGround->setFrictionAll(friction);
    }
    else if (name == "absolute" || name == "relative")
    {
        Uint16 x = n->getUnsignedProperty("x");
        Uint16 y = n->getUnsignedProperty("y");
        Uint16 w = n->getUnsignedProperty("w");
        Uint16 h = n->getUnsignedProperty("h");

        int gx = n->getIntProperty("gx", 0);
        int gy = n->getIntProperty("gy", 0);

        int vx = n->getIntProperty("vx", 0);
        int vy = n->getIntProperty("vy", 0);

        unsigned friction = n->getUnsignedProperty("friction", 0);

        if (name == "absolute")
        {
            for (Uint16 yy = y; yy < y + h; yy++)
            {
                for (Uint16 xx = x; xx < x + w; xx++)
                {
                    playGround->setXGravity(xx, yy, gx);
                    playGround->setYGravity(xx, yy, gy);
                    playGround->setXVelocity(xx, yy, vx);
                    playGround->setYVelocity(xx, yy, vy);
                    playGround->setFriction(xx, yy, friction);
                }
            }
        }
        else
        {
            for (Uint16 yy = y; yy < y + h; yy++)
            {
                for (Uint16 xx = x; xx < x + w; xx++)
                {
                    playGround->addXGravity(xx, yy, gx);
                    playGround->addYGravity(xx, yy, gy);
                    playGround->addXVelocity(xx, yy, vx);
                    playGround->addYVelocity(xx, yy, vy);
                    playGround->addFriction(xx, yy, friction);
                }
            }
        }
    }
    else
    {
        throw XMLException(
            std::string("The gravity node '").append(name)
            .append("' is unknown"));
    }
}


//----------------------------------------------------------------------------
void CreateDecorationsVisitor::do_visit(const XMLNode *n)
{
    if (n->getName() == "decorations")
    {
        n->acceptAllChilds(*this);
    }
    else
    {
        ObjectRepository::getInstance()->addObject(DecorationBase::create(n));
    }
}


//----------------------------------------------------------------------------
PlayGround::PlayGround()
{
    m_collisionSurface = NULL;
    m_mapSurface = NULL;
    m_shadowSurface = NULL;

    memset(&m_boundingBox, 0, sizeof(m_boundingBox));
}

//----------------------------------------------------------------------------
PlayGround::~PlayGround()
{
    ZAP_SURFACE(m_collisionSurface);
    ZAP_SURFACE(m_mapSurface);
    ZAP_SURFACE(m_shadowSurface);
}

//----------------------------------------------------------------------------
void PlayGround::init(const char *mission, const XMLNode *playGroundNode)
{
    destroy();

    sm_instance = new PlayGround();
    sm_instance->readMapFrom(
        mission, playGroundNode->getStringProperty("map").c_str());

    const XMLNode *gravity = playGroundNode->getNode("gravity");
    if (gravity)
    {
        CreateGravityVisitor v1;
        gravity->accept(v1);
    }

    CreateDecorationsVisitor v2;
    playGroundNode->getMandatoryNode("decorations")->accept(v2);
}

//----------------------------------------------------------------------------
void PlayGround::destroy()
{
    ZAP_POINTER(sm_instance);
}


//----------------------------------------------------------------------------
Crate *PlayGround::hasCrateFor(const Ship *ship) const
{
    HasCrateForShipConstVisitor v(ship);
    ObjectRepository::getInstance()->acceptObjects(v);
    return const_cast<Crate*>(v.getCrate());
}

//----------------------------------------------------------------------------
void PlayGround::resize(Uint16 xTiles, Uint16 yTiles)
{
    m_boundingBox.w = xTiles * 16;
    m_boundingBox.h = yTiles * 16;

    m_collisionSurface = SDL_CALLS::CreateRGBSurface(
        SDL_SWSURFACE, m_boundingBox.w, m_boundingBox.h, 32);
    m_mapSurface = SDL_CALLS::CreateRGBSurface(
        SDL_SWSURFACE, m_boundingBox.w, m_boundingBox.h, 32);
    m_shadowSurface = SDL_CALLS::CreateRGBSurface(
        SDL_SWSURFACE, m_boundingBox.w, m_boundingBox.h, 32);

    m_background.resize(xTiles, yTiles, true);
    m_xGravity.resize(xTiles, yTiles, 0);
    m_yGravity.resize(xTiles, yTiles, 100);
    m_xVelocity.resize(xTiles, yTiles, 0);
    m_yVelocity.resize(xTiles, yTiles, 0);
    m_friction.resize(xTiles, yTiles, 50);
}


//----------------------------------------------------------------------------
void PlayGround::readMapFrom(const char *mission, const char *map)
{
    std::string full;
    full.append(Configuration::getInstance()->getDataDir())
        .append("/levels/").append(mission).append("/").append(map);

    File f(full.c_str(), "r");

    Uint8 version = f.readUint8();
    switch (version)
    {
    case 0x01:
        readMapV1(f);
        break;

    case 0x02:
        readMapV2(f);
        break;

    default:
    {
        std::ostringstream s;
        s << "The map file version 0x"
          << std::setfill('0') << std::setw(2) << std::hex << version
          << " is not supported";

        throw Exception(s.str());
    }
    }
}

//----------------------------------------------------------------------------
void PlayGround::readMapV1(File &f)
{
    Uint16 xTiles = f.readUint16();
    Uint16 yTiles = f.readUint16();
    resize(xTiles, yTiles);

    Uint8 layers = f.readUint8();
    if (layers == 0)
    {
        throw Exception("There must be at least one layer");
    }

    do
    {
        Uint8 category = f.readUint8();

        for (Uint16 y=0; y<yTiles; y++)
        {
            SDL_Rect r;
            for (Uint16 x=0; x<xTiles; x++)
            {
                r.x = 16*x;
                r.y = 16*y;

                Uint8 tile = f.readUint8();

                if (tile == 0x00)  continue;

                Uint8 alpha = (category == 0xff ? 0x80 : SDL_ALPHA_OPAQUE);
                const SDL_Surface *tileSurface =
                    Tiles::getInstance()->getTile(category, tile, alpha, 0x80);

                SDL_CALLS::BlitSurface(tileSurface, NULL, m_mapSurface, &r);

                if (category != 0x00 && category != 0xff)
                {
                    SDL_CALLS::BlitSurface(
                        tileSurface, NULL, m_collisionSurface, &r);

                    m_background.set(x, y, false);
                }
            }
        }
    }
    while (--layers > 0);

    SDL_CALLS::BlitSurface(m_mapSurface, NULL, m_shadowSurface, NULL);
}

//----------------------------------------------------------------------------
void PlayGround::readMapV2(File &f)
{
    Uint16 xTiles = f.readUint16();
    Uint16 yTiles = f.readUint16();
    resize(xTiles, yTiles);

    Uint8 layers = f.readUint8();
    if (layers == 0)
    {
        throw Exception("There must be at least one layer");
    }

    do
    {
        Uint8 category = f.readUint8();
        Uint8 background = f.readUint8();
        Uint8 alpha = f.readUint8();
        Uint8 brightness = f.readUint8();

        if (category == 0x00 || category == 0xff)
        {
            std::ostringstream s;
            s << "The category index 0x"
              << std::setfill('0') << std::setw(2) << std::hex << category
              << " shall not be used for version 2 maps";

            throw Exception(s.str());
        }

        for (Uint16 y=0; y<yTiles; y++)
        {
            SDL_Rect r;
            for (Uint16 x=0; x<xTiles; x++)
            {
                r.x = 16*x;
                r.y = 16*y;

                Uint8 tile = f.readUint8();

                if (tile == 0x00)  continue;

                const SDL_Surface *tileSurface =
                    Tiles::getInstance()->getTile(
                        category, tile, alpha, brightness);

                SDL_CALLS::BlitSurface(tileSurface, NULL, m_mapSurface, &r);

                if (background == 0x00)
                {
                    SDL_CALLS::BlitSurface(
                        tileSurface, NULL, m_collisionSurface, &r);

                    m_background.set(x, y, false);
                }
            }
        }
    }
    while (--layers > 0);

    SDL_CALLS::BlitSurface(m_mapSurface, NULL, m_shadowSurface, NULL);
}


//----------------------------------------------------------------------------
template <class T>
static T getTileAverage(const Matrix<T> &matrix, const SDL_Rect &r,
                        const T oobValue)
{
    // If r lies (partially or fully) outside of the given matrix,
    // we take oobValue for all outside points.

    // Since the matrix value within a tile is always constant,
    // we don't need to sweep through pixels.
    // Instead, we sweep through the tiles.

    T sum = 0;
    Sint16 rxrw = r.x + r.w;
    Sint16 ryrh = r.y + r.h;

    Sint16 w = 0, h = 0;

    Sint16 y = r.y;
    while (y < ryrh)
    {
        Sint16 x = r.x;
        while (x < rxrw)
        {
            Sint16 xTile = x/16;
            Sint16 yTile = y/16;

            w = 16 - x % 16;
            if (w > 16)  w -= 16;
            if (x + w > rxrw)
            {
                w = rxrw - x;
            }

            h = 16 - y % 16;
            if (h > 16)  h -= 16;
            if (y + h > ryrh)
            {
                h = ryrh - y;
            }

            if (xTile < 0 || yTile < 0 ||
                (unsigned)xTile >= matrix.getXSize() ||
                (unsigned)yTile >= matrix.getYSize())
            {
                xTile = 0;
                yTile = 0;
            }

            sum += matrix.get(xTile, yTile) * w * h;

            x += w;
        }

        y += h;
    }

    return r.w == 0 || r.h == 0 ? oobValue : sum / (r.w * r.h);
}

//----------------------------------------------------------------------------
int PlayGround::getXGravity(const SDL_Rect &r) const
{
    return getTileAverage(m_xGravity, r, m_xGravity.get(0, 0));
}

//----------------------------------------------------------------------------
int PlayGround::getYGravity(const SDL_Rect &r) const
{
    return getTileAverage(m_yGravity, r, m_yGravity.get(0, 0));
}

//----------------------------------------------------------------------------
int PlayGround::getXVelocity(const SDL_Rect &r) const
{
    return getTileAverage(m_xVelocity, r, m_xVelocity.get(0, 0));
}

//----------------------------------------------------------------------------
int PlayGround::getYVelocity(const SDL_Rect &r) const
{
    return getTileAverage(m_yVelocity, r, m_yVelocity.get(0, 0));
}

//----------------------------------------------------------------------------
unsigned PlayGround::getFriction(const SDL_Rect &r) const
{
    return getTileAverage(m_friction, r, (unsigned)0);
}


//----------------------------------------------------------------------------
bool PlayGround::collidesWith(const ObjectBase *object) const
{
    const SDL_Rect &objPos = object->getPosition();

    // Catch objects lying fully or partially outside of the playground.
    SDL_Rect tmpPos = { 0, 0, 0, 0 };
    if (!SDL_TOOLS::intersect(getBoundingBox(), objPos, tmpPos) ||
        !SDL_TOOLS::equal(objPos, tmpPos))
    {
        return true;
    }

    Sint16 objPosX = objPos.x;
    Sint16 objPosY = objPos.y;

    // The most common case will be, that the object does not collide with the
    // playfield. So we check, if the object's surface only overlaps background
    // tiles. If yes, we can return false, without checking the pixel data
    // of the object and the playground.

    Sint16 yMax = (objPosY + objPos.h + 15) / 16;
    Sint16 xMax = (objPosX + objPos.w + 15) / 16;

    for (Sint16 y = objPosY/16; y < yMax; y++)
    {
        for (Sint16 x = objPosX/16; x < xMax; x++)
        {
            if (!isBackgroundTile(x, y))
            {
                // @todo Since the object usually lies on some
                // background tiles, it would be faster to perform
                // a pixel check only for the object tiles,
                // that do not lie on a background tile.
                // However, this shouldn't be much of a bottleneck...

                return SDL_TOOLS::isCollision(
                    object->getSurface(), objPos,
                    getCollisionSurface(), getBoundingBox());
            }
        }
    }

    return false;
}

//----------------------------------------------------------------------------
bool PlayGround::collidesWithParticle(const ParticleBase *particle) const
{
    const SDL_Rect &pos = particle->getPosition();

    return SDL_TOOLS::inside(pos.x, pos.y, m_boundingBox) ?
        SDL_TOOLS::getPixel(getCollisionSurface(), pos.x, pos.y) != 0 : true;
}

//----------------------------------------------------------------------------
const Platform *PlayGround::isInLandingZone(const Ship *ship) const
{
    FindLandingPlatformConstVisitor v(ship);
    ObjectRepository::getInstance()->acceptObjects(v);
    return v.getPlatform();
}

//----------------------------------------------------------------------------
bool PlayGround::isBackgroundBetween(Uint16 x1, Uint16 y1,
                                     Uint16 x2, Uint16 y2) const
{
    Sint16 diffX = (Sint16)x1 - (Sint16)x2;
    Sint16 diffY = (Sint16)y1 - (Sint16)y2;

    if (diffX == 0 && diffY == 0)
    {
        return isBackgroundTile(x1, y1);
    }

    if (abs(diffX) > abs(diffY))
    {
        // Iterate through the x axis.

        if (x1 < x2)
        {
            return do_isBackgroundBetweenX(x1, y1, x2, y2);
        }
        else
        {
            return do_isBackgroundBetweenX(x2, y2, x1, y1);
        }
    }
    else
    {
        // Iterate through the y axis.

        if (y1 < y2)
        {
            return do_isBackgroundBetweenY(x1, y1, x2, y2);
        }
        else
        {
            return do_isBackgroundBetweenY(x2, y2, x1, y1);
        }
    }
}

//----------------------------------------------------------------------------
bool PlayGround::do_isBackgroundBetweenX(Uint16 x1, Uint16 y1,
                                         Uint16 x2, Uint16 y2) const
{
    double k = (1.0 * (y2-y1)) / (1.0 * (x2-x1));
    double y = 1.0 * y1;

    for (Uint16 x = x1; x <= x2; x++)
    {
        if (!isBackgroundTile(x, (Uint16)round(y)))
        {
            return false;
        }

        y += k;
    }

    return true;
}

//----------------------------------------------------------------------------
bool PlayGround::do_isBackgroundBetweenY(Uint16 x1, Uint16 y1,
                                         Uint16 x2, Uint16 y2) const
{
    double k = (1.0 * (x2-x1)) / (1.0 * (y2-y1));
    double x = 1.0 * x1;

    for (Uint16 y = y1; y <= y2; y++)
    {
        if (!isBackgroundTile((Uint16)round(x), y))
        {
            return false;
        }

        x += k;
    }

    return true;
}

//----------------------------------------------------------------------------
void PlayGround::update()
{
    ObjectRepository *r = ObjectRepository::getInstance();

    m_updateRects.clear();

    OverdrawObjectsConstVisitor v1(m_mapSurface, m_shadowSurface);
    r->acceptAll(v1);

    UpdateObjectsVisitor v2;
    r->acceptAll(v2);

    CollisionDetectorVisitor v3;
    r->acceptAll(v3);

    r->removeObjectsToRemove();

    BlitObjectsVisitor v4(m_shadowSurface);
    r->acceptAll(v4);
}

//----------------------------------------------------------------------------
void PlayGround::updateForPreview()
{
    BlitObjectsVisitor v(m_shadowSurface);
    ObjectRepository::getInstance()->acceptObjects(v);
}


//----------------------------------------------------------------------------
void OverdrawObjectsConstVisitor::do_visit(const ParticleBase *p)
{
    // This is an optimized version of do_do_visit(p);

    const SDL_Rect &pos = p->getPosition();

    assert(SDL_TOOLS::inside(pos.x, pos.y, m_mapSurface));

    Uint8 *mapPixel = (Uint8*)m_mapSurface->pixels
        + pos.y * m_mapSurface->pitch + pos.x * 4;
    Uint8 *shadowPixel = (Uint8*)m_shadowSurface->pixels
        + pos.y * m_shadowSurface->pitch + pos.x * 4;

    *(Uint32*)shadowPixel = *(Uint32*)mapPixel;
}

//----------------------------------------------------------------------------
void OverdrawObjectsConstVisitor::do_do_visit(const ObjectBase *o)
{
    const SDL_Rect *r = o->getPositionPointer();
    SDL_CALLS::BlitSurface(m_mapSurface, r, m_shadowSurface, r);
}



//----------------------------------------------------------------------------
void UpdateObjectsVisitor::do_visit(Grenade *g)
{
    do_do_visit(g);

    if (g->isExplode())
    {
        // The GameControl's collision handling does what we want,
        // though we actually have no collision here.
        GameControlBase::getInstance()->onCollision(g);
    }
}

//----------------------------------------------------------------------------
void UpdateObjectsVisitor::do_visit(Missile *m)
{
    do_do_visit(m);

    if (m->isExplode())
    {
        // The GameControl's collision handling does what we want,
        // though we actually have no collision here.
        GameControlBase::getInstance()->onCollision(m);
    }
}

//----------------------------------------------------------------------------
void UpdateObjectsVisitor::do_visit(ParticleBase *p)
{
    do_do_visit(p);
}

//----------------------------------------------------------------------------
void UpdateObjectsVisitor::do_visit(Ship *s)
{
    do_do_visit(s);

    if (s->isExplode())
    {
        // The GameControl's collision handling does what we want,
        // though we actually have no collision here.
        GameControlBase::getInstance()->onCollision(s);
    }
}

//----------------------------------------------------------------------------
void UpdateObjectsVisitor::do_do_visit(ObjectBase *o)
{
    SDL_Rect r = o->getPosition();
    o->update();
    SDL_TOOLS::unite(r, o->getPosition(), r);

    PlayGround::getInstance()->addUpdateRect(r);
}



//----------------------------------------------------------------------------
void CollisionDetectorVisitor::ParticleCollisionDetectorVisitor::do_onCollision(ObjectBase *o)
{
    if (ObjectBase::isCollision(m_particle, o))
    {
        m_particle->setHidden(true);
        setCollision();
    }
}


//----------------------------------------------------------------------------
void CollisionDetectorVisitor::ShipCollisionDetectorVisitor::do_visit(Grenade *g)
{
    if (ObjectBase::isCollision(m_ship, g))
    {
        GameControlBase::getInstance()->onCollision(m_ship, g);
        setCollision();
    }
}

//----------------------------------------------------------------------------
void CollisionDetectorVisitor::ShipCollisionDetectorVisitor::do_visit(Missile *m)
{
    if (ObjectBase::isCollision(m_ship, m))
    {
        GameControlBase::getInstance()->onCollision(m_ship, m);
        setCollision();
    }
}

//----------------------------------------------------------------------------
void CollisionDetectorVisitor::ShipCollisionDetectorVisitor::do_visit(MortarBase *m)
{
    if (ObjectBase::isCollision(m_ship, m))
    {
        GameControlBase::getInstance()->onCollision(m_ship, m);
        setCollision();
    }
}

//----------------------------------------------------------------------------
void CollisionDetectorVisitor::ShipCollisionDetectorVisitor::do_visit(ProjectileBase *p)
{
    if (ObjectBase::isCollision(m_ship, p))
    {
        GameControlBase::getInstance()->onCollision(m_ship, p);
        setCollision();
    }
}

//----------------------------------------------------------------------------
void CollisionDetectorVisitor::ShipCollisionDetectorVisitor::do_visit(SAMBatteryBase *s)
{
    if (ObjectBase::isCollision(m_ship, s))
    {
        GameControlBase::getInstance()->onCollision(m_ship, s);
        setCollision();
    }
}

//----------------------------------------------------------------------------
void CollisionDetectorVisitor::ShipCollisionDetectorVisitor::do_visit(Tank *t)
{
    if (ObjectBase::isCollision(m_ship, t))
    {
        GameControlBase::getInstance()->onCollision(m_ship, t);
        setCollision();
    }
}

//----------------------------------------------------------------------------
void CollisionDetectorVisitor::ShipCollisionDetectorVisitor::do_visit(TurretBase *t)
{
    if (ObjectBase::isCollision(m_ship, t))
    {
        GameControlBase::getInstance()->onCollision(m_ship, t);
        setCollision();
    }
}

//----------------------------------------------------------------------------
void CollisionDetectorVisitor::ShipCollisionDetectorVisitor::do_onCollision(ObjectBase *object)
{
    if (ObjectBase::isCollision(m_ship, object))
    {
        GameControlBase::getInstance()->onCollision(m_ship);
        setCollision();
    }
}


//----------------------------------------------------------------------------
void CollisionDetectorVisitor::MissileCollisionDetectorVisitor::do_visit(Grenade *g)
{
    if (ObjectBase::isCollision(m_missile, g))
    {
        GameControlBase::getInstance()->onCollision(m_missile, g);
        setCollision();
    }
}

//----------------------------------------------------------------------------
void CollisionDetectorVisitor::MissileCollisionDetectorVisitor::do_visit(MortarBase *m)
{
    if (ObjectBase::isCollision(m_missile, m))
    {
        GameControlBase::getInstance()->onCollision(m_missile, m);
        setCollision();
    }
}

//----------------------------------------------------------------------------
void CollisionDetectorVisitor::MissileCollisionDetectorVisitor::do_visit(ProjectileBase *p)
{
    if (ObjectBase::isCollision(m_missile, p))
    {
        GameControlBase::getInstance()->onCollision(m_missile, p);
        setCollision();
    }
}

//----------------------------------------------------------------------------
void CollisionDetectorVisitor::MissileCollisionDetectorVisitor::do_visit(SAMBatteryBase *s)
{
    if (ObjectBase::isCollision(m_missile, s))
    {
        GameControlBase::getInstance()->onCollision(m_missile, s);
        setCollision();
    }
}

//----------------------------------------------------------------------------
void CollisionDetectorVisitor::MissileCollisionDetectorVisitor::do_visit(SwitchBase *s)
{
    if (ObjectBase::isCollision(m_missile, s))
    {
        GameControlBase::getInstance()->onCollision(m_missile, s);
        setCollision();
    }
}

//----------------------------------------------------------------------------
void CollisionDetectorVisitor::MissileCollisionDetectorVisitor::do_visit(TurretBase *t)
{
    if (ObjectBase::isCollision(m_missile, t))
    {
        GameControlBase::getInstance()->onCollision(m_missile, t);
        setCollision();
    }
}

//----------------------------------------------------------------------------
void CollisionDetectorVisitor::MissileCollisionDetectorVisitor::do_visit(Tank *t)
{
    if (ObjectBase::isCollision(m_missile, t))
    {
        GameControlBase::getInstance()->onCollision(m_missile, t);
        setCollision();
    }
}

//----------------------------------------------------------------------------
void CollisionDetectorVisitor::MissileCollisionDetectorVisitor::do_onCollision(ObjectBase *o)
{
    if (ObjectBase::isCollision(m_missile, o))
    {
        GameControlBase::getInstance()->onCollision(m_missile);
        setCollision();
    }
}


//----------------------------------------------------------------------------
void CollisionDetectorVisitor::GrenadeCollisionDetectorVisitor::do_visit(MortarBase *m)
{
    if (ObjectBase::isCollision(m_grenade, m))
    {
        GameControlBase::getInstance()->onCollision(m_grenade, m);
        setCollision();
    }
}

//----------------------------------------------------------------------------
void CollisionDetectorVisitor::GrenadeCollisionDetectorVisitor::do_visit(SwitchBase *s)
{
    if (ObjectBase::isCollision(m_grenade, s))
    {
        GameControlBase::getInstance()->onCollision(m_grenade, s);
        setCollision();
    }
}

//----------------------------------------------------------------------------
void CollisionDetectorVisitor::GrenadeCollisionDetectorVisitor::do_visit(Tank *t)
{
    if (ObjectBase::isCollision(m_grenade, t))
    {
        GameControlBase::getInstance()->onCollision(m_grenade, t);
        setCollision();
    }
}

//----------------------------------------------------------------------------
void CollisionDetectorVisitor::GrenadeCollisionDetectorVisitor::do_visit(TurretBase *t)
{
    if (ObjectBase::isCollision(m_grenade, t))
    {
        GameControlBase::getInstance()->onCollision(m_grenade, t);
        setCollision();
    }
}

//----------------------------------------------------------------------------
void CollisionDetectorVisitor::GrenadeCollisionDetectorVisitor::do_onCollision(ObjectBase *o)
{
    if (ObjectBase::isCollision(m_grenade, o))
    {
        GameControlBase::getInstance()->onCollision(m_grenade);
        setCollision();
    }
}


//----------------------------------------------------------------------------
void CollisionDetectorVisitor::ProjectileCollisionDetectorVisitor::do_visit(MortarBase *m)
{
    if (ObjectBase::isCollision(m_projectile, m))
    {
        GameControlBase::getInstance()->onCollision(m_projectile, m);
        setCollision();
    }
}

//----------------------------------------------------------------------------
void CollisionDetectorVisitor::ProjectileCollisionDetectorVisitor::do_visit(SAMBatteryBase *s)
{
    if (ObjectBase::isCollision(m_projectile, s))
    {
        GameControlBase::getInstance()->onCollision(m_projectile, s);
        setCollision();
    }
}

//----------------------------------------------------------------------------
void CollisionDetectorVisitor::ProjectileCollisionDetectorVisitor::do_visit(SwitchBase *s)
{
    if (ObjectBase::isCollision(m_projectile, s))
    {
        GameControlBase::getInstance()->onCollision(m_projectile, s);
        setCollision();
    }
}

//----------------------------------------------------------------------------
void CollisionDetectorVisitor::ProjectileCollisionDetectorVisitor::do_visit(Tank *t)
{
    if (ObjectBase::isCollision(m_projectile, t))
    {
        GameControlBase::getInstance()->onCollision(m_projectile, t);
        setCollision();
    }
}

//----------------------------------------------------------------------------
void CollisionDetectorVisitor::ProjectileCollisionDetectorVisitor::do_visit(TurretBase *t)
{
    if (ObjectBase::isCollision(m_projectile, t))
    {
        GameControlBase::getInstance()->onCollision(m_projectile, t);
        setCollision();
    }
}

//----------------------------------------------------------------------------
void CollisionDetectorVisitor::ProjectileCollisionDetectorVisitor::do_onCollision(ObjectBase *o)
{
    if (ObjectBase::isCollision(m_projectile, o))
    {
        GameControlBase::getInstance()->onCollision(m_projectile);
        setCollision();
    }
}


//----------------------------------------------------------------------------
void CollisionDetectorVisitor::do_visit(ParticleBase *p)
{
    PlayGround *playGround = PlayGround::getInstance();

    if (playGround->collidesWithParticle(p))
    {
        p->setHidden(true);
    }
    else
    {
        ParticleCollisionDetectorVisitor v(p);
        ObjectRepository::getInstance()->acceptObjects(v);
    }
}

//----------------------------------------------------------------------------
void CollisionDetectorVisitor::do_visit(Ship *s)
{
    PlayGround *playGround = PlayGround::getInstance();

    if (playGround->collidesWith(s))
    {
        GameControlBase::getInstance()->onCollision(s);
    }
    else
    {
        ShipCollisionDetectorVisitor v(s);
        ObjectRepository::getInstance()->acceptObjects(v);
    }
}

//----------------------------------------------------------------------------
void CollisionDetectorVisitor::do_visit(Missile *m)
{
    PlayGround *playGround = PlayGround::getInstance();

    if (playGround->collidesWith(m))
    {
        GameControlBase::getInstance()->onCollision(m);
    }
    else
    {
        MissileCollisionDetectorVisitor v(m);
        ObjectRepository::getInstance()->acceptObjects(v);
    }
}

//----------------------------------------------------------------------------
void CollisionDetectorVisitor::do_visit(Grenade *g)
{
    PlayGround *playGround = PlayGround::getInstance();

    if (playGround->collidesWith(g))
    {
        GameControlBase::getInstance()->onCollision(g);
    }
    else
    {
        GrenadeCollisionDetectorVisitor v(g);
        ObjectRepository::getInstance()->acceptObjects(v);
    }
}

//----------------------------------------------------------------------------
void CollisionDetectorVisitor::do_visit(ProjectileBase *p)
{
    PlayGround *playGround = PlayGround::getInstance();

    if (playGround->collidesWith(p))
    {
        GameControlBase::getInstance()->onCollision(p);
    }
    else
    {
        ProjectileCollisionDetectorVisitor v(p);
        ObjectRepository::getInstance()->acceptObjects(v);
    }
}


//----------------------------------------------------------------------------
void BlitObjectsVisitor::do_visit(ParticleBase *p)
{
    // This is an optimized version of do_do_visit(p);

    p->updateSurface();

    const SDL_Rect &pos = p->getPosition();
    const SDL_Surface *surface = p->getSurface();

    assert(SDL_TOOLS::inside(pos.x, pos.y, m_shadowSurface));

    Uint8 *shadowPixel = (Uint8*)m_shadowSurface->pixels
        + pos.y * m_shadowSurface->pitch + pos.x * 4;

    *(Uint32*)shadowPixel = *(Uint32*)surface->pixels;
}

//----------------------------------------------------------------------------
void BlitObjectsVisitor::do_do_visit(ObjectBase *o)
{
    o->updateSurface();
    o->blitTo(m_shadowSurface);
}
